from dataclasses import dataclass, field
from typing import List, Dict, Any

# the implemented measurements for 
valid_measurements = {
    'tide',
    'wind',
    'water_temp',
    'air_temp',
    'air_press',
    'wave',
    'conductivity',
}

@dataclass
class Result:
    """Simplified requests.Response returned from RestAdapter

    Args:
        status_code (int): HTTP Status
        message (str, optional): Human readable result. Defaults to ''.
        data (List[Dict], optional): Python list of Dicts. Defaults to None.
    """
    status_code: int
    message: str = ''
    data: List[Dict] = field(default_factory=list)
    
    def __post__init__(self, data):
        self.data = data if data else [] #init data if dne

@dataclass
class Station:
    id: str
    lat: float
    lon: float
    api: Any
    tide: bool = False
    wind: bool = False
    water_temp: bool = False
    air_temp: bool = False
    air_press: bool = False
    wave: bool = False
    conductivity: bool = False
    is_active: bool = False
    name: str = ''
    
    def __post_init__(self):
        # toggle if self.is_active
        if len(self.supported_measurements()) > 0:
            self.is_active = True
            
    def supported_measurements(self) -> List[str]:
        """Return list of supported measurements for station

        Returns:
            List[str]: _description_
        """
        # 
        supported = []
        for key in valid_measurements:
            if self.__getattribute__(key):
                supported.append(key)
        return supported
        
    def is_supported(self, value:str) -> bool:
        # Implemented this to always run comparison on lower case casting of strings
        for x in self.supported_measurements():
            if value in x:
                return True
        return False
        
        
if __name__ == '__main__':
    # Result tests
    test = Result(200)
    
    # Station Tests
    test = Station()
    # test internal methods


import logging
from datetime import datetime
from typing import Dict, List

from seastate.api.rest_adapter import RestAdapter
from seastate.exceptions import SeaStateException

class TidesAndCurrentsApi:
    def __init__(self, api_key: str = '', ssl_verify: bool = True, logger: logging.Logger = None):
        """Constructor for TidesAndCurrentsApi, composed with RestAdapter 

        Args:
            api_key (str, optional): Not used here. Defaults to ''.
            ssl_verify (bool, optional): Defaults to True.
            logger (logging.Logger, optional): Pass explictly else will be created with __name__.
        """
        self._logger = logger or logging.getLogger(__name__)
        self._rest_adapter = RestAdapter('api.tidesandcurrents.noaa.gov', api_key, ssl_verify, logger)
        
    def measurements_from_date_range(self, measurement:str, station_id:str, start:datetime, end: datetime) -> List[Dict]:
        """Returns Result for station ID and datetime start and end. Args should be validated externally

        Args:
            measurement (str): measurement type can be tide|wind|water_temp|air_temp|air_press|conductivity|tide_prediction.
            station_id (str): Station ID.
            start (datetime): start datetime.
            end (datetime): end datetime.

        Raises:
            SeaStateException: _description_

        Returns:
            Result: _description_
        """
        
        # each measurement type will have a different endpoint "product" param and key to unpack response
        # always test str on lowercase casting
        measurement = measurement.lower()
        if 'tide' in measurement:
            product = 'water_level'
            measurement_key = 'data'
        elif 'wind' in measurement:
            product = 'wind'
            measurement_key = 'data'
        elif 'water_temp' in measurement:
            product = 'water_temperature'
            measurement_key = 'data'
        elif 'air_temp' in measurement:
            product = 'air_temperature'
            measurement_key = 'data'
        elif 'air_press' in measurement:
            product = 'air_pressure'
            measurement_key = 'data'
        elif 'wave' in measurement:
            raise SeaStateException("Unsupported measurement requested")
        elif 'conductivity' in measurement:
            product = 'conductivity'
            measurement_key = 'data'
        else:
            raise SeaStateException("Unsupported measurement requested")

        # formatting datetime to endpoint param
        begin_date = f"{start.year}{start.month:02}{start.day:02}"
        end_date = f"{end.year}{end.month:02}{end.day:02}"
        
        ep_params = {
            'begin_date': begin_date,
            'end_date': end_date,
            'station': str(station_id),
            'product': product,
            'interval': 'h',
            'datum': 'MTL',
            'units': 'metric',
            'time_zone': 'lst_ldt',
            'application': 'seastate',
            'format': 'json',
        }
        
        # Call endpoint
        endpoint='api/prod/datagetter?'
        result = self._rest_adapter.get(endpoint=endpoint,ep_params=ep_params)
        
        # unpack to return specified measurement
        # since TidesAndCurrents returns 1 product per endpoint, no need to parse, just unpack Json
        # details here: https://api.tidesandcurrents.noaa.gov/api/prod/responseHelp.html
        # 
        try:
            data = result.data[measurement_key]
        except (KeyError) as e:
            self._logger.error(result.data)
            raise SeaStateException("TidesAndCurrentsApi unpacking error") from e
        
        if len(data) == 0:
            raise SeaStateException("No data retrieved")
        
        return data

        
if __name__ == '__main__':
    api = TidesAndCurrentsApi()
    result = api.hourly('wind',9410230, datetime.today(),datetime.today())
    print(result)
    result = api.hourly(9410230, datetime.today(),datetime.today(),'air_temp')
    result = api.hourly(9410230, datetime.today(),datetime.today(),'water_temp')


from dataclasses import field
import logging

from seastate.api.datasources import DataSources
from seastate.exceptions import SeaStateException
from seastate.models import Station
from seastate.utils import haversine

class ApiMediator:
    def __init__(self, measurement: str, lat: float, lon: float, exclude: list = field(default_factory=list), ssl_verify: bool = True, logger: logging.Logger = None):
        """Constructor for configured api endpoint based on measurement type, gps coordinates and exclude filters 
        """
        self._ssl_verify = ssl_verify
        self._logger = logger or logging.getLogger(__name__)
        self._target_lat = lat
        self._target_lon = lon
        self.exclude = exclude
        self.measurement = measurement
        self.station = self.nearest_station() # nearest_station handles the selection logic
        self.api = self.station.api
        self.distance = haversine(self._target_lat,self._target_lon, self.station.lat, self.station.lon)
        
    ## Properties
    @property
    def _target_lat(self):
        return self.__target_lat
    
    @_target_lat.setter
    def _target_lat(self,value):
        #  Latitudes are in range of -90 to 90
        if -90 <= value <= 90:
            self.__target_lat = value
        else:
            raise SeaStateException("Latitude must be between -90 and 90 degrees")
    
    @property
    def _target_lon(self):
        return self.__target_lon
    
    @_target_lon.setter
    def _target_lon(self,value):
        #  Longitudes are in range of -180 to 180
        if -180 <= value <= 180:
            self.__target_lon = value
        else:
            raise SeaStateException("Longitude must be between -180 and 180 degrees")

    # # Methods
    def nearest_station(self) -> Station:
        # if measurement
        # Return dict of stations
        # n<2000 so just return them all
        try:
            # Grab all stations
            stations = DataSources().all()
        except (KeyError) as e:
            raise SeaStateException("Error retrieving stations") from e

        # Find station closest to input coordinates using haversine
        # and is active for specified measurement
        min_ptr = float('inf') # Initialize minimum pointer
        for eval_station in stations:
            # skip if station is in exclude list
            if eval_station.id in self.exclude:
                continue
            # skip if eval_station is inactive
            if not eval_station.is_active:
                continue
            # skip if station does not support measurement
            if not eval_station.is_supported(self.measurement):
                continue
            # compute distance between SeaState target coords and current station
            new_val = haversine(self._target_lat, self._target_lon, eval_station.lat, eval_station.lon)
            # if closer, update new minimum
            if new_val < min_ptr:
                min_ptr = new_val
                station = eval_station
        return station



if __name__ == '__main__':
    api = ApiMediator('Tide',32,-117) # testing San Diego for tide
    # Station should have tide as a valid measurement
    assert api.station.tide
    assert api.station.is_supported(api.measurement)
    # out of bound lat/lon should throw errors
    
    # testing wind
    api = ApiMediator('Wind',32,-117) # testing San Diego for tide




import logging
from typing import List

import defusedxml.minidom
from seastate.api.noaa_ndbc import NdbcApi
from seastate.api.noaa_tidesandcurrents import TidesAndCurrentsApi
from seastate.api.rest_adapter import RestAdapter
from seastate.exceptions import SeaStateException
from seastate.models import Station


class DataSources:
    def __init__(self, logger: logging.Logger = None):
        self._logger = logger or logging.getLogger(__name__)
        
    def all(self) -> List[Station]:
        """Convenience function for all stations

        Returns:
            List[Station]: _description_
        """
        stations = self.ndbc_stations()
        stations += self.tides_and_currents_stations()
        return stations
    
    def ndbc_stations(self) -> List[Station]:
        """Parse a text table

        Returns:
            List[Station]: _description_
        """
        # todo: refactor this spaghetti code
        # check station status by retrieving latest measurements from all stations
        result = RestAdapter('www.ndbc.noaa.gov/').get('data/latest_obs/latest_obs.txt')
        
        # loop through text lines and test if station is ouputting specified measurement
        stations = []
        for i, value in enumerate(result.data.split('\n')):
            # skipping 2 header lines
            if i == 0 or i == 1:
                continue #skip header and unit lines
            
            # split current line by delimiter and clear empty values
            line = value.split(' ')    
            while ' ' in line:
                line.remove(' ')
            while '' in line:
                line.remove('')
            
            # skip corrupted lines
            if len(line) != 22:
                self._logger.info("No station info in this line: " + str(line))
                continue

            # parse stationID, gps, and confirm active measurement sources
            # 'MM' is NOAA's notation for missing measurement
            temp_station = {
                'id': line[0],
                'lat': float(line[1]),
                'lon': float(line[2]),
                'api': NdbcApi(),
                'tide': True if line[21] != 'MM' else False,
                'wind': True if line[8] != 'MM' else False,
                'water_temp': True if line[18] != 'MM' else False,
                'air_temp': True if line[17] != 'MM' else False,
                'air_press': True if line[15] != 'MM' else False,
                'wave': True if line[11] != 'MM' else False,
            }
            
            # construct station with temp var
            stations.append(Station(**temp_station))
        
        # return list of stations
        return stations

    def tides_and_currents_stations(self) -> List[Station]:
        """Parse xml doc

        Raises:
            SeaStateException: _description_

        Returns:
            List[Station]: _description_
        """
        # todo: refactor this spaghetti code
        # TidesAndCurrent station capabilities are kept here
        result = RestAdapter('opendap.co-ops.nos.noaa.gov/').get('stations/stationsXML.jsp')
        
        # parse xml elements safely with defusedxml -> []
        dom = defusedxml.minidom.parseString(result.data)
        station_elements = dom.getElementsByTagName('station')
        
        # traverse station elements to build Station objects
        stations = []
        for line in station_elements:
            temp_station = {}
            # parse station metadata into temporary dict
            try:
                temp_station['name'] = line.getAttribute('name')
                temp_station['id'] = line.getAttribute('ID')
                temp_station['lat'] = float(line.getElementsByTagName('lat')[0].firstChild.nodeValue)
                temp_station['lon'] = float(line.getElementsByTagName('long')[0].firstChild.nodeValue)
                temp_station['api'] = TidesAndCurrentsApi()
            except (IndexError) as e:
                # Faulty station, skip station node
                self._logger.warn(e + str(line))
                continue
            
            # separate loop to parse individual measurements
            for m in line.getElementsByTagName('parameter'):
                name = m.attributes['name'].value
                status = True if m.attributes['status'].value == '1' else False
                if 'Water Level' in name and status:
                    temp_station['tide'] = True
                elif 'Winds' in name and status:
                    temp_station['wind'] = True
                    # todo: wind dir and gust may or may not be true
                elif 'Air Temp' in name and status:
                    temp_station['air_temp'] = True
                elif 'Water Temp' in name and status:
                    temp_station['water_temp'] = True
                elif 'Air Pressure' in name and status:
                    temp_station['air_press'] = True
                elif 'Conductivity' in name and status:
                    temp_station['conductivity'] = True

            # construct station with temp var
            stations.append(Station(**temp_station))
            
        # Check parsing was succesful
        if len(stations) == 0:
            raise SeaStateException("No stations successfully parsed, please submit issue")
        return stations

if __name__ == '__main__':
    tnc = DataSources().tides_and_currents_stations()
    ndbc = DataSources().ndbc_stations()    


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['seastate', 'seastate.api']

package_data = \
{'': ['*']}

install_requires = \
['defusedxml>=0.7.1,<0.8.0',
 'numpy>=1.23.1,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'urllib3>=1.26.10,<2.0.0']

setup_kwargs = {
    'name': 'seastate',
    'version': '0.1.2',
    'description': 'Collect ocean measurement data based on location and timeframe',
    'long_description': "# Seastate\n\n## Summary\nCollect ocean measurement data based on location and timeframe\n\n**Features**\n- Closest active station is selected for each measurement\n- Reaches into secondary historical archives when required\n- Returns pandas dataframe compatible lists\n- Available measurements: Tide, wind, water temp, air temp, air pressure, conductivity and swell information\n- Datasources: NOAA NDBC, NOAA Tides and Currents\n\n## Installing\n`pip install seastate`\n\n## Quick start\n```\nfrom seastate.seastate import SeaState\nfrom datetime import datetime\n\n# make SeaState object for specific location\nsan_diego = SeaState(32,-117)\n\n# retrieve measurements for today\nsan_diego_today = san_diego.measurements_from_date_range(datetime.today())\n\nsan_diego_today['tide'] -> [{t: time, v: value, s: stdev]\nsan_diego_today['wind']-> [{t: time, v: value, d: direction, g: gust}]\nsan_diego_today['water_temp']-> [{t: time, v: value}]\nsan_diego_today['air_temp']-> [{t: time, v: value}]\nsan_diego_today['air_press']-> [{t: time, v: value}]\nsan_diego_today['wave']-> [{t: time, v: Wave Height, dpd: Dominant Period, mwd: dpd Direction}]\nsan_diego_today['conductivity']-> [{t: time, v: value}] \n```\nMeasurement details for NDBC are [here](https://www.ndbc.noaa.gov/measdes.shtml) and for Tides and Currents [here](https://api.tidesandcurrents.noaa.gov/api/prod/responseHelp.html)\n\n## Measurement x API breakdown\n| Measurement | T&C | NDBC |\n| ---: | :---: | :---: |\n|tide | y | y |\n|wind | y | y |\n|water_temp | y | y |\n|air_temp | y | y |\n|air_press | y | y |\n|wave |  | y |\n|conductivity | y |  |\n\n## More Examples\n### Measurements for past 30 days\n```\nfrom seastate.seastate import SeaState\nfrom datetime import datetime, timedelta\n\nstart = datetime.today()-timedelta(days=30)\nend = datetime.today()\nsan_diego_past_30 = san_diego.measurements_from_date_range(start,end)\n```\n### Hourly Slices\n```\nsan_diego_past_30_hourly = san_diego.hourly(start,end) # this returns a single reading per hour\n# experimental feature\n# no guarantee between APIs that readings will align or exist in all cases\n```\n",
    'author': 'Ivo',
    'author_email': 'ivorivetta@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ivorivetta/seastate',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



import logging
from datetime import datetime, timedelta
from typing import Dict, List

from seastate.api.rest_adapter import RestAdapter
from seastate.exceptions import SeaStateException

class NdbcApi:
    def __init__(self, api_key: str = '', ssl_verify: bool = True, logger: logging.Logger = None):
        """Constructor for NdbcApi, composed with RestAdapter 

        Args:
            api_key (str, optional): Not used here. Defaults to ''.
            ssl_verify (bool, optional): Defaults to True.
            logger (logging.Logger, optional): Pass explictly else will be created with __name__.
        """
        self._logger = logger or logging.getLogger(__name__)
        self._rest_adapter = RestAdapter('www.ndbc.noaa.gov', api_key, ssl_verify, logger)
        
    def measurements_from_date_range(self, measurement:str, station_id:str, start:datetime, end: datetime) -> List[Dict]:
        """Returns Result for stationID and datetime start and end

        Args:
            station (str): station ID.
            start (datetime): start datetime.
            end (datetime): end datetime.

        Raises:
            SeaStateException: _description_

        Returns:
            Result: _description_
        """
        # todo refactor this spaghetti code
        measurement = measurement.lower()
        if 'tide' in measurement:
            measurement_key = ['TIDE']
        elif 'wind' in measurement:
            measurement_key = ['WSPD','SPD']
            # additional wind information handled when unpacking
        elif 'water_temp' in measurement:
            measurement_key = ['WTMP']
        elif 'air_temp' in measurement:
            measurement_key = ['ATMP']
        elif 'air_press' in measurement:
            measurement_key = ['PRES','BARO','BAR']
        elif 'wave' in measurement:
            measurement_key = ['WVHT','H0']
            # additional swell information handled when unpacking
        elif 'conductivity' in measurement:
            raise SeaStateException("Unsupported measurement requested, please report issue")
        else:
            raise SeaStateException("Unsupported measurement requested, please report issue")
        
        # Collect endpoints needed to cover daterange
        endpoints = []
        # daterange spans realtime endpoint (45 days)
        archive_cutoff_date = datetime.today() - timedelta(days=45)
        if start > archive_cutoff_date or end > archive_cutoff_date:
            # "realtime" endpoint covers prev 45 days
            endpoints += [f"data/realtime2/{station_id}.txt"]
        # daterange spans archive, for current year
        # cutoff month has a unique endpoint
        if start < archive_cutoff_date < end:
            requested_months = [archive_cutoff_date.strftime("%b")]
            endpoints += [f"data/stdmet/{x}/{station_id}.txt" for x in requested_months]
        # within current year, endpoints are grouped by month
        if start < archive_cutoff_date and end.year == datetime.today().year:
            # use January if start year precedes the archive cutoff year
            # else use start month
            start_range = 1 if start.year < archive_cutoff_date.year else start.month
            # end of range will be up to cutoff
            # duplicates to be filtered downstream
            end_range = archive_cutoff_date.month
            # Endpoint uses month number and abbreviated month designations (%b)
            requested_months = [(str(x), datetime(1,x,1).strftime("%b")) for x in range(start_range,end_range)]
            endpoints += [f"view_text_file.php?filename={station_id}{x}{datetime.today().year}.txt.gz&dir=data/stdmet/{y}/" for x, y in requested_months]
        # daterange spans prior years
        # 
        if start < archive_cutoff_date and start.year < datetime.today().year:
            # if end year is this year, handled upstream
            # else, it is a prior year and needs to be included
            end_range = end.year if end.year == datetime.today().year else end.year + 1
            requested_years = [str(x) for x in range(start.year,end_range)]
            endpoints += [f"view_text_file.php?filename={station_id}h{x}.txt.gz&dir=data/historical/stdmet/" for x in requested_years]

        result = ''
        for endpoint in endpoints:
            try:
                res = self._rest_adapter.get(endpoint=endpoint)
            except (SeaStateException) as e:
                self._logger.warning(f"{str(e)} for {endpoint}")
                continue
            result += res.data
                
        
        # unpack text result
        # realtime ndbc reports are a text file with previous 45 days of measurements
        # archival are a years worth
        # todo: unpack text file into t: and v:
        data = []
        for i, value in enumerate(result.split('\n')):
            # turn str to list
            # known anomaly: 2 space characters sometimes delimiting 
            line = value.replace('  ', ' ').split(' ')
            while ' ' in line:
                line.remove(' ') #remove blanks
            while '' in line:
                line.remove('') #remove blanks
            
            # handle corrupted lines
            if len(line) == 0:
                self._logger.info(f"ignoring line #{i}:{str(line)}")
                continue
            
            # handle header lines
            if line[0] in ['#YY', 'YY']:
                # header row starts with #YY in realtime, YY in archive
                header = line
                # strip '#' from '#YY'
                header[0] = header[0].strip('#')
                continue
            elif line[0] in ['#yr']:
                # skipping units line
                continue

            
            # ignore lines outside of daterange
            # archive files sometimes use 95 instead of 1995
            if not (start.year <= int(line[0]) <= end.year) and not (abs(start.year % 100) <= int(line[0]) <= abs(end.year % 100)):
                # 4 digit year representation
                continue
            if not start.month <= int(line[1]) <= end.month:
                continue
            if not start.day <= int(line[2]) <= end.day:
                continue
            
            # line is in requested daterange
            # parse timestamp
            timestamp = datetime(
                year=int(line[0] if len(line[0])==4 else '19' + line[0]), #rebuild dates with archive format
                month=int(line[1]),
                day=int(line[2]),
                hour=int(line[3]),
                minute=int(line[4] if 'mm' in header else 0) # archival format is hourly
                ).isoformat(sep= ' ')
            # parse measurement column
            # because of changes in the source api column names over the years
            # we try to unpack with the possible variants
            # details here: https://www.ndbc.noaa.gov/measdes.shtml
            temp_data = {}
            # unpack time and main value
            for key in measurement_key:
                try:
                    temp_data['t'] = timestamp
                    temp_data['v'] = line[header.index(key)]
                except (KeyError, ValueError) as e:
                    pass
            # check for success before continuing
            if len(temp_data) < 2:
                self._logger.error(value)
                raise SeaStateException("NdbcApi unpacking error, please report issue")
            # wind sometimes has direction and gust data
            if 'wind' in measurement:
                # for wind direction:
                for key in ['WDIR', 'WD','DIR']:
                    try:
                        temp = line[header.index(key)]
                        temp_data['d'] = temp if temp and '999' not in temp and 'MM' not in temp else None # 999 for direction
                    except Exception as e:
                        self._logger.debug(f"{e} for {measurement}")
                # For wind gust:
                for key in ['GST', 'GSP']:
                    try:
                        temp = line[header.index(key)]
                        temp_data['g'] = temp if temp and '99' not in temp and 'MM' not in temp else None # 99 for decimal
                    except Exception as e:
                        self._logger.debug(f"{e} for {measurement}")
            # wave sometimes has period and direction
            if 'wave' in measurement:
                # for dominant wave period:
                for key in ['DPD','DOMPD']:
                    try:
                        temp = line[header.index(key)]
                        temp_data['dpd'] = temp if temp and '99' not in temp and 'MM' not in temp else None  # 99 for decimal
                    except Exception as e:
                        self._logger.debug(f"{e} for {measurement}")
                # For dominant wave direction:
                for key in ['MWD']:
                    try:
                        temp = line[header.index(key)]
                        temp_data['mwd'] = temp if temp and '999' not in temp and 'MM' not in temp else None # 999 for direction
                    except Exception as e:
                        self._logger.debug(f"{e} for {measurement}")
                # For average wave period
                for key in ['APD','AVP']:
                    try:
                        temp = line[header.index(key)]
                        temp_data['apd'] = temp if temp and '99' not in temp and 'MM' not in temp else None
                    except Exception as e:
                        self._logger.debug(f"{e} for {measurement}")
            data.append(temp_data)
            
        # log warning if no data recovered for daterange
        if len(data) == 0:
            self._logger.warning(f"No {measurement} data recovered for daterange: {str(start.date())} : {str(end.date())}")

        # todo: scrub duplicates between cutoff month and realtime
        return data

        
if __name__ == '__main__':
    api = NdbcApi()
    # check daterange within realtime
    # result = api.hourly('wind',46224,datetime.today(),datetime.today()-timedelta(days=2))
    
    # check daterange for request only in prior years
    # result = api.hourly('wind',46224,datetime.today()-timedelta(days=2*365+30),datetime.today()-timedelta(days=1*365+30))

    # check daterange spanning archive and realtime, into prior year
    # result = api.hourly('wind',46224,datetime.today()-timedelta(days=365+30),datetime.today())
    
    # check old archive with different headers
    # check date format
    result = api.hourly('air_press',42040,datetime(1996,2,1),datetime(1996,2,2))



import logging
from datetime import datetime, timedelta
from typing import Dict, Union

from seastate.api.api_mediator import ApiMediator

class SeaState:
    def __init__(self, lat: float, lon: float, exclude: list=[], logger: logging.Logger = None):
        self._logger = logger or logging.getLogger(__name__)
        self.lat = float(lat)
        self.lon = float(lon)
        self.exclude = list(exclude)
        self.tide = ApiMediator('tide', self.lat, self.lon, self.exclude)
        self.wind = ApiMediator('wind', self.lat, self.lon, self.exclude)
        self.water_temp = ApiMediator('water_temp', self.lat, self.lon, self.exclude)
        self.air_temp = ApiMediator('air_temp', self.lat, self.lon, self.exclude)
        self.air_press = ApiMediator('air_press', self.lat, self.lon, self.exclude)
        self.wave = ApiMediator('wave', self.lat, self.lon, self.exclude)
        self.conductivity = ApiMediator('conductivity', self.lat, self.lon, self.exclude)
        
    def measurements_from_date_range(self, start: datetime = None, end: Union[datetime, timedelta] = None) -> Dict:
        # Process timeframe
        # todo: handle iso date strings
        # todo: handle bad date strings
        if start and end: # all values provided, handle endpoint
            # end can be declared relative to start as a timedelta
            if isinstance(end, timedelta):
                end = start + end
        elif start and not end: # no end provided, return same day as start
            end = start
        elif not start and not end: # nothing provided, return today
            start = datetime.today()
            end = datetime.today()
            
        # remove microseconds to pass comparison tests
        # remove hours and minutes from start
        # set hours and minutes to end of day for start
        if isinstance(start,datetime):
            start = start.replace(microsecond=0)
            start = start.replace(hour=0, minute=0)
        if isinstance(end,datetime):
            end = end.replace(microsecond=0)
            end = end.replace(hour=23, minute=0)
            
        # log warning if end is before start
        if end < start:
            self._logger.warning("end is after start")
        
        # Get data
        data = {}
        #todo: have a class property that summarizes the active apis for this list
        for api_key in ['tide', 'wind', 'water_temp', 'air_temp', 'air_press', 'wave', 'conductivity']:
            # access the configured api to generate response with hourly method 
            data[api_key] = self.__getattribute__(api_key).api.measurements_from_date_range(
                measurement = self.__getattribute__(api_key).measurement,
                station_id = self.__getattribute__(api_key).station.id,
                start = start,
                end = end)
        return data

    def hourly(self, start: datetime = None, end: Union[datetime, timedelta] = None) -> Dict:
        """Convenience method to return 1 sample per hour for each api

        Args:
            start (datetime, optional): _description_. Defaults to None.
            end (Union[datetime, timedelta], optional): _description_. Defaults to None.

        Returns:
            Dict: _description_
        """
        data_all = self.measurements_from_date_range(start,end)
        data= {}
        for key in data_all:
            # find the first minute within the hour
            # the apis typically use the same minute 
            first_minute = 0
            while datetime.fromisoformat(data_all[key][0]['t']).minute != first_minute:
                first_minute += 1
                # prevent endless loop 
                if first_minute == 60:
                    data[key] = []
                    break
            # keep one reading per hour
            data[key] = [x for x in data_all[key] if datetime.fromisoformat(x['t']).minute == first_minute]
        return data
    
if __name__ == '__main__':
    api = SeaState(32,-117)
    start = datetime(2022,7,5)
    end = datetime(2022,7,5)
    a = api.hourly(start,end)
    
    # check daterange within realtime
    # result = api.hourly(datetime.today()-timedelta(days=2),datetime.today())
    
    # # check daterange for request only in prior years
    # result = api.hourly(datetime.today()-timedelta(days=2*365+30),datetime.today()-timedelta(days=1*365+30))

    # # check daterange spanning archive and realtime, into prior year
    # result = api.hourly(datetime.today()-timedelta(days=365+30),datetime.today())
    
    # # check old archive with different headers
    # # check date format
    # result = api.hourly('air_press',42040,datetime(1996,2,1),datetime(1996,2,2))

